@@ -64,15 +64,20 @@ typedef struct {
 
 	GSList      *connections;      /* CONTAINS: SoupConnection */
 	guint        num_conns;
+
+	guint        num_messages;
 } SoupSessionHost;
 
 typedef struct {
 	char *ssl_ca_file;
 	SoupSSLCredentials *ssl_creds;
+	gboolean ssl_strict;
 
 	SoupMessageQueue *queue;
 
 	char *user_agent;
+	char *accept_language;
+	gboolean accept_language_auto;
 
 	GSList *features;
 	GHashTable *features_cache;
@@ -102,6 +107,8 @@ static void queue_message   (SoupSession *session, SoupMessage *msg,
 static void requeue_message (SoupSession *session, SoupMessage *msg);
 static void cancel_message  (SoupSession *session, SoupMessage *msg,
 			     guint status_code);
+static void auth_required   (SoupSession *session, SoupMessage *msg,
+			     SoupAuth *auth, gboolean retrying);
 
 static void auth_manager_authenticate (SoupAuthManager *manager,
 				       SoupMessage *msg, SoupAuth *auth,
@@ -119,6 +126,8 @@ enum {
 	REQUEST_STARTED,
 	REQUEST_UNQUEUED,
 	AUTHENTICATE,
+	CONNECTION_CREATED,
+	TUNNELING,
 	LAST_SIGNAL
 };
 
@@ -132,9 +141,12 @@ enum {
 	PROP_MAX_CONNS_PER_HOST,
 	PROP_USE_NTLM,
 	PROP_SSL_CA_FILE,
+	PROP_SSL_STRICT,
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_USER_AGENT,
+	PROP_ACCEPT_LANGUAGE,
+	PROP_ACCEPT_LANGUAGE_AUTO,
 	PROP_IDLE_TIMEOUT,
 	PROP_ADD_FEATURE,
 	PROP_ADD_FEATURE_BY_TYPE,
@@ -181,6 +193,8 @@ soup_session_init (SoupSession *session)
 	 * so hold a ref on the default GResolver.
 	 */
 	priv->resolver = g_resolver_get_default ();
+
+	priv->ssl_strict = TRUE;
 }
 
 static void
@@ -210,7 +224,10 @@ finalize (GObject *object)
 	g_hash_table_destroy (priv->conns);
 
 	g_free (priv->user_agent);
+	g_free (priv->accept_language);
 
+	if (priv->ssl_ca_file)
+		g_free (priv->ssl_ca_file);
 	if (priv->ssl_creds)
 		soup_ssl_free_client_credentials (priv->ssl_creds);
 
@@ -235,6 +252,7 @@ soup_session_class_init (SoupSessionClass *session_class)
 	session_class->queue_message = queue_message;
 	session_class->requeue_message = requeue_message;
 	session_class->cancel_message = cancel_message;
+	session_class->auth_required = auth_required;
 
 	/* virtual method override */
 	object_class->dispose = dispose;
@@ -375,6 +393,33 @@ soup_session_class_init (SoupSessionClass *session_class)
 			      SOUP_TYPE_AUTH,
 			      G_TYPE_BOOLEAN);
 
+	signals[CONNECTION_CREATED] =
+		g_signal_new ("connection-created",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0,
+			      NULL, NULL,
+			      soup_marshal_NONE__OBJECT,
+			      G_TYPE_NONE, 1,
+			      /* SoupConnection is private, so we can't use
+			       * SOUP_TYPE_CONNECTION here.
+			       */
+			      G_TYPE_OBJECT);
+
+	signals[TUNNELING] =
+		g_signal_new ("tunneling",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0,
+			      NULL, NULL,
+			      soup_marshal_NONE__OBJECT,
+			      G_TYPE_NONE, 1,
+			      /* SoupConnection is private, so we can't use
+			       * SOUP_TYPE_CONNECTION here.
+			       */
+			      G_TYPE_OBJECT);
+
+
 	/* properties */
 	/**
 	 * SOUP_SESSION_PROXY_URI:
@@ -468,6 +513,24 @@ soup_session_class_init (SoupSessionClass *session_class)
 				     "File containing SSL CA certificates",
 				     NULL,
 				     G_PARAM_READWRITE));
+	/**
+	 * SOUP_SESSION_SSL_STRICT:
+	 *
+	 * Alias for the #SoupSession:ignore-ssl-cert-errors
+	 * property. By default, when validating certificates against
+	 * a CA file, Soup will consider invalid certificates as a
+	 * connection error. Setting this property to %TRUE makes soup
+	 * ignore the errors, and make the connection.
+	 *
+	 * Since: 2.30
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_SSL_STRICT,
+		g_param_spec_boolean (SOUP_SESSION_SSL_STRICT,
+				      "Strictly validate SSL certificates",
+				      "Whether certificate errors should be considered a connection error",
+				      TRUE,
+				      G_PARAM_READWRITE));
 	/**
 	 * SOUP_SESSION_ASYNC_CONTEXT:
 	 *
@@ -534,6 +597,55 @@ soup_session_class_init (SoupSessionClass *session_class)
 				     NULL,
 				     G_PARAM_READWRITE));
 
+	/**
+	 * SoupSession:accept-language:
+	 *
+	 * If non-%NULL, the value to use for the "Accept-Language" header
+	 * on #SoupMessage<!-- -->s sent from this session.
+	 *
+	 * Setting this will disable
+	 * #SoupSession:accept-language-auto.
+	 *
+	 * Since: 2.30
+	 **/
+	/**
+	 * SOUP_SESSION_ACCEPT_LANGUAGE:
+	 *
+	 * Alias for the #SoupSession:accept-language property, qv.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_ACCEPT_LANGUAGE,
+		g_param_spec_string (SOUP_SESSION_ACCEPT_LANGUAGE,
+				     "Accept-Language string",
+				     "Accept-Language string",
+				     NULL,
+				     G_PARAM_READWRITE));
+
+	/**
+	 * SoupSession:accept-language-auto:
+	 *
+	 * If %TRUE, #SoupSession will automatically set the string
+	 * for the "Accept-Language" header on every #SoupMessage
+	 * sent, based on the return value of g_get_language_names().
+	 *
+	 * Setting this will override any previous value of
+	 * #SoupSession:accept-language.
+	 *
+	 * Since: 2.30
+	 **/
+	/**
+	 * SOUP_SESSION_ACCEPT_LANGUAGE_AUTO:
+	 *
+	 * Alias for the #SoupSession:accept-language-auto property, qv.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_ACCEPT_LANGUAGE_AUTO,
+		g_param_spec_boolean (SOUP_SESSION_ACCEPT_LANGUAGE_AUTO,
+				      "Accept-Language automatic mode",
+				      "Accept-Language automatic mode",
+				      FALSE,
+				      G_PARAM_READWRITE));
+
 	/**
 	 * SoupSession:add-feature:
 	 *
@@ -618,6 +730,88 @@ safe_str_equal (const char *a, const char *b)
 	return strcmp (a, b) == 0;
 }
 
+/* Converts a language in POSIX format and to be RFC2616 compliant    */
+/* Based on code from epiphany-webkit (ephy_langs_append_languages()) */
+static gchar *
+posix_lang_to_rfc2616 (const gchar *language)
+{
+	/* Don't include charset variants, etc */
+	if (strchr (language, '.') || strchr (language, '@'))
+		return NULL;
+
+	/* Ignore "C" locale, which g_get_language_names() always
+	 * includes as a fallback.
+	 */
+	if (!strcmp (language, "C"))
+		return NULL;
+
+	return g_strdelimit (g_ascii_strdown (language, -1), "_", '-');
+}
+
+/* Converts @quality from 0-100 to 0.0-1.0 and appends to @str */
+static gchar *
+add_quality_value (const gchar *str, int quality)
+{
+	g_return_val_if_fail (str != NULL, NULL);
+
+	if (quality > 0 && quality < 100) {
+		double qvalue = quality / 100.0;
+		return g_strdup_printf ("%s;q=%.2g", str, qvalue);
+	} else {
+		/* Just dup the string in this case */
+		return g_strdup (str);
+	}
+}
+
+/* Returns a RFC2616 compliant languages list from system locales */
+static gchar *
+accept_languages_from_system (void)
+{
+	const char * const * lang_names;
+	GPtrArray *langs = NULL;
+	char *lang, **langs_array, *langs_str;
+	int delta;
+	int i;
+
+	lang_names = g_get_language_names ();
+	g_return_val_if_fail (lang_names != NULL, NULL);
+
+	/* Build the array of languages */
+	langs = g_ptr_array_new ();
+	for (i = 0; lang_names[i] != NULL; i++) {
+		lang = posix_lang_to_rfc2616 (lang_names[i]);
+		if (lang)
+			g_ptr_array_add (langs, lang);
+	}
+
+	/* Add quality values */
+	if (langs->len < 10)
+		delta = 10;
+	else if (langs->len < 20)
+		delta = 5;
+	else
+		delta = 1;
+
+	for (i = 0; i < langs->len; i++) {
+		lang = langs->pdata[i];
+		langs->pdata[i] = add_quality_value (lang, 100 - i * delta);
+		g_free (lang);
+	}
+
+	/* Fallback: add "en" if list is empty */
+	if (langs->len == 0)
+		g_ptr_array_add (langs, g_strdup ("en"));
+
+	g_ptr_array_add (langs, NULL);
+	langs_array = (char **)langs->pdata;
+	langs_str = g_strjoinv (", ", langs_array);
+
+	g_strfreev (langs_array);
+	g_ptr_array_free (langs, FALSE);
+
+	return langs_str;
+}
+
 static void
 set_property (GObject *object, guint prop_id,
 	      const GValue *value, GParamSpec *pspec)
@@ -672,6 +866,9 @@ set_property (GObject *object, guint prop_id,
 			priv->ssl_creds = NULL;
 		}
 
+		break;
+	case PROP_SSL_STRICT:
+		priv->ssl_strict = g_value_get_boolean (value);
 		break;
 	case PROP_ASYNC_CONTEXT:
 		priv->async_context = g_value_get_pointer (value);
@@ -696,6 +893,22 @@ set_property (GObject *object, guint prop_id,
 		} else
 			priv->user_agent = g_strdup (user_agent);
 		break;
+	case PROP_ACCEPT_LANGUAGE:
+		g_free (priv->accept_language);
+		priv->accept_language = g_strdup (g_value_get_string (value));
+		priv->accept_language_auto = FALSE;
+		break;
+	case PROP_ACCEPT_LANGUAGE_AUTO:
+		priv->accept_language_auto = g_value_get_boolean (value);
+		if (priv->accept_language) {
+			g_free (priv->accept_language);
+			priv->accept_language = NULL;
+		}
+
+		/* Get languages from system if needed */
+		if (priv->accept_language_auto)
+			priv->accept_language = accept_languages_from_system ();
+		break;
 	case PROP_IDLE_TIMEOUT:
 		priv->idle_timeout = g_value_get_uint (value);
 		break;
@@ -750,6 +963,9 @@ get_property (GObject *object, guint prop_id,
 	case PROP_SSL_CA_FILE:
 		g_value_set_string (value, priv->ssl_ca_file);
 		break;
+	case PROP_SSL_STRICT:
+		g_value_set_boolean (value, priv->ssl_strict);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
@@ -759,6 +975,12 @@ get_property (GObject *object, guint prop_id,
 	case PROP_USER_AGENT:
 		g_value_set_string (value, priv->user_agent);
 		break;
+	case PROP_ACCEPT_LANGUAGE:
+		g_value_set_string (value, priv->accept_language);
+		break;
+	case PROP_ACCEPT_LANGUAGE_AUTO:
+		g_value_set_boolean (value, priv->accept_language_auto);
+		break;
 	case PROP_IDLE_TIMEOUT:
 		g_value_set_uint (value, priv->idle_timeout);
 		break;
@@ -804,16 +1026,12 @@ soup_session_host_new (SoupSession *session, SoupURI *uri)
 	return host;
 }
 
-/* Note: get_host_for_message doesn't lock the host_lock. The caller
- * must do it itself if there's a chance the host doesn't already
- * exist.
- */
+/* Requires host_lock to be locked */
 static SoupSessionHost *
-get_host_for_message (SoupSession *session, SoupMessage *msg)
+get_host_for_uri (SoupSession *session, SoupURI *uri)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupSessionHost *host;
-	SoupURI *uri = soup_message_get_uri (msg);
 
 	host = g_hash_table_lookup (priv->hosts, uri);
 	if (host)
@@ -825,6 +1043,16 @@ get_host_for_message (SoupSession *session, SoupMessage *msg)
 	return host;
 }
 
+/* Note: get_host_for_message doesn't lock the host_lock. The caller
+ * must do it itself if there's a chance the host doesn't already
+ * exist.
+ */
+static SoupSessionHost *
+get_host_for_message (SoupSession *session, SoupMessage *msg)
+{
+	return get_host_for_uri (session, soup_message_get_uri (msg));
+}
+
 static void
 free_host (SoupSessionHost *host)
 {
@@ -840,12 +1068,20 @@ free_host (SoupSessionHost *host)
 	g_slice_free (SoupSessionHost, host);
 }	
 
+static void
+auth_required (SoupSession *session, SoupMessage *msg,
+	       SoupAuth *auth, gboolean retrying)
+{
+	g_signal_emit (session, signals[AUTHENTICATE], 0, msg, auth, retrying);
+}
+
 static void
 auth_manager_authenticate (SoupAuthManager *manager, SoupMessage *msg,
 			   SoupAuth *auth, gboolean retrying,
 			   gpointer session)
 {
-	g_signal_emit (session, signals[AUTHENTICATE], 0, msg, auth, retrying);
+	SOUP_SESSION_GET_CLASS (session)->auth_required (
+		session, msg, auth, retrying);
 }
 
 #define SOUP_METHOD_IS_SAFE(method) (method == SOUP_METHOD_GET || \
@@ -869,10 +1105,12 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 	     !SOUP_METHOD_IS_SAFE (msg->method)) ||
 	    (msg->status_code == SOUP_STATUS_MOVED_PERMANENTLY &&
 	     msg->method == SOUP_METHOD_POST)) {
-		/* Redirect using a GET */
-		g_object_set (msg,
-			      SOUP_MESSAGE_METHOD, SOUP_METHOD_GET,
-			      NULL);
+		if (msg->method != SOUP_METHOD_HEAD) {
+			/* Redirect using a GET */
+			g_object_set (msg,
+				      SOUP_MESSAGE_METHOD, SOUP_METHOD_GET,
+				      NULL);
+		}
 		soup_message_set_request (msg, NULL,
 					  SOUP_MEMORY_STATIC, NULL, 0);
 		soup_message_headers_set_encoding (msg->request_headers,
@@ -909,6 +1147,8 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 	 */
 	new_uri = soup_uri_new_with_base (soup_message_get_uri (msg), new_loc);
 	if (!new_uri || !new_uri->host) {
+		if (new_uri)
+			soup_uri_free (new_uri);
 		soup_message_set_status_full (msg,
 					      SOUP_STATUS_MALFORMED,
 					      "Invalid Redirect URL");
@@ -941,44 +1181,48 @@ soup_session_send_queue_item (SoupSession *session,
 					      "User-Agent", priv->user_agent);
 	}
 
+	if (priv->accept_language &&
+	    !soup_message_headers_get_list (item->msg->request_headers,
+					    "Accept-Language")) {
+		soup_message_headers_append (item->msg->request_headers,
+					     "Accept-Language",
+					     priv->accept_language);
+	}
+
 	g_signal_emit (session, signals[REQUEST_STARTED], 0,
 		       item->msg, soup_connection_get_socket (conn));
 	soup_connection_send_request (conn, item->msg);
 }
 
 gboolean
-soup_session_try_prune_connection (SoupSession *session)
+soup_session_cleanup_connections (SoupSession *session,
+				  gboolean     prune_idle)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
-	GPtrArray *conns;
+	GSList *conns = NULL, *c;
 	GHashTableIter iter;
 	gpointer conn, host;
-	int i;
-
-	conns = g_ptr_array_new ();
+	SoupConnectionState state;
 
 	g_mutex_lock (priv->host_lock);
 	g_hash_table_iter_init (&iter, priv->conns);
 	while (g_hash_table_iter_next (&iter, &conn, &host)) {
-		/* Don't prune a connection that is currently in use,
-		 * or hasn't been used yet.
-		 */
-		if (soup_connection_get_state (conn) == SOUP_CONNECTION_IDLE &&
-		    soup_connection_last_used (conn) > 0)
-			g_ptr_array_add (conns, g_object_ref (conn));
+		state = soup_connection_get_state (conn);
+		if (state == SOUP_CONNECTION_REMOTE_DISCONNECTED ||
+		    (prune_idle && state == SOUP_CONNECTION_IDLE))
+			conns = g_slist_prepend (conns, g_object_ref (conn));
 	}
 	g_mutex_unlock (priv->host_lock);
 
-	if (!conns->len) {
-		g_ptr_array_free (conns, TRUE);
+	if (!conns)
 		return FALSE;
-	}
 
-	for (i = 0; i < conns->len; i++) {
-		soup_connection_disconnect (conns->pdata[i]);
-		g_object_unref (conns->pdata[i]);
+	for (c = conns; c; c = c->next) {
+		conn = c->data;
+		soup_connection_disconnect (conn);
+		g_object_unref (conn);
 	}
-	g_ptr_array_free (conns, TRUE);
+	g_slist_free (conns);
 
 	return TRUE;
 }
@@ -1018,36 +1262,33 @@ soup_session_connection_failed (SoupSession *session,
 	g_mutex_lock (priv->host_lock);
 	host = g_hash_table_lookup (priv->conns, conn);
 	g_mutex_unlock (priv->host_lock);
-	if (!host)
-		return;
 
-	connection_disconnected (conn, session);
-
-	if (host->connections) {
-		/* Something went wrong this time, but we have at
-		 * least one open connection to this host. So just
-		 * leave the message in the queue so it can use that
-		 * connection once it's free.
-		 */
-		return;
+	if (host)
+		connection_disconnected (conn, session);
+	else {
+		host = g_object_get_data (G_OBJECT (conn), "SoupSessionHost");
+		if (!host)
+			return;
 	}
 
-	/* Assume that there's something wrong with the host, and
-	 * cancel any other messages waiting for a connection to it,
+	/* Cancel any other messages waiting for a connection to it,
 	 * since they're out of luck.
 	 */
 	g_object_ref (session);
 	for (item = soup_message_queue_first (priv->queue); item; item = soup_message_queue_next (priv->queue, item)) {
 		msg = item->msg;
-		if (get_host_for_message (session, msg) == host)
+		if (SOUP_MESSAGE_IS_STARTING (msg) &&
+		    get_host_for_message (session, msg) == host)
 			soup_session_cancel_message (session, msg, status);
 	}
 	g_object_unref (session);
 }
 
 static void
-tunnel_connected (SoupSession *session, SoupMessage *msg, gpointer user_data)
+tunnel_connected (SoupMessage *msg, gpointer user_data)
 {
+	SoupSession *session = user_data;
+
 	if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
 		SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 		SoupMessageQueueItem *item =
@@ -1059,6 +1300,7 @@ tunnel_connected (SoupSession *session, SoupMessage *msg, gpointer user_data)
 		g_object_set (item->conn,
 			      SOUP_CONNECTION_PROXY_URI, NULL,
 			      NULL);
+		soup_message_queue_item_unref (item);
 	}
 }
 
@@ -1069,6 +1311,7 @@ soup_session_make_connect_message (SoupSession *session,
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupURI *uri;
 	SoupMessage *msg;
+	SoupMessageQueueItem *item;
 
 	uri = soup_uri_new (NULL);
 	soup_uri_set_scheme (uri, SOUP_URI_SCHEME_HTTPS);
@@ -1081,10 +1324,16 @@ soup_session_make_connect_message (SoupSession *session,
 
 	/* Call the base implementation of soup_session_queue_message
 	 * directly, to add msg to the SoupMessageQueue and cause all
-	 * the right signals to be emitted.
+	 * the right signals to be emitted. We can't use
+	 * queue_message's callback arg in this case because that's
+	 * actually implemented by the subclasses.
 	 */
-	queue_message (session, msg, tunnel_connected, NULL);
-	return soup_message_queue_lookup (priv->queue, msg);
+	g_signal_connect (msg, "finished",
+			  G_CALLBACK (tunnel_connected), session);
+	queue_message (session, msg, NULL, NULL);
+	item = soup_message_queue_lookup (priv->queue, msg);
+	g_object_unref (msg);
+	return item;
 }
 
 /**
@@ -1133,7 +1382,7 @@ soup_session_get_connection (SoupSession *session,
 	SoupAddress *remote_addr, *tunnel_addr;
 	SoupSSLCredentials *ssl_creds;
 	GSList *conns;
-	gboolean has_pending = FALSE;
+	int num_pending = 0;
 	SoupURI *uri;
 
 	g_mutex_lock (priv->host_lock);
@@ -1145,14 +1394,13 @@ soup_session_get_connection (SoupSession *session,
 			g_mutex_unlock (priv->host_lock);
 			return conns->data;
 		} else if (soup_connection_get_state (conns->data) == SOUP_CONNECTION_CONNECTING)
-			has_pending = TRUE;
+			num_pending++;
 	}
 
-	if (has_pending) {
-		/* We've already started one connection to this
-		 * address, so don't start another one until it's
-		 * done.
-		 */
+	/* Limit the number of pending connections; num_messages / 2
+	 * is somewhat arbitrary...
+	 */
+	if (num_pending > host->num_messages / 2) {
 		g_mutex_unlock (priv->host_lock);
 		return NULL;
 	}
@@ -1192,14 +1440,18 @@ soup_session_get_connection (SoupSession *session,
 		SOUP_CONNECTION_TUNNEL_ADDRESS, tunnel_addr,
 		SOUP_CONNECTION_PROXY_URI, item->proxy_uri,
 		SOUP_CONNECTION_SSL_CREDENTIALS, ssl_creds,
+		SOUP_CONNECTION_SSL_STRICT, priv->ssl_strict,
 		SOUP_CONNECTION_ASYNC_CONTEXT, priv->async_context,
 		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
 		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
 		NULL);
+	g_object_set_data (G_OBJECT (conn), "SoupSessionHost", host);
 	g_signal_connect (conn, "disconnected",
 			  G_CALLBACK (connection_disconnected),
 			  session);
 
+	g_signal_emit (session, signals[CONNECTION_CREATED], 0, conn);
+
 	g_hash_table_insert (priv->conns, conn, host);
 
 	priv->num_conns++;
@@ -1224,6 +1476,7 @@ message_finished (SoupMessage *msg, gpointer user_data)
 	SoupMessageQueueItem *item = user_data;
 	SoupSession *session = item->session;
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupSessionHost *host;
 
 	if (item->conn) {
 		g_object_unref (item->conn);
@@ -1232,6 +1485,12 @@ message_finished (SoupMessage *msg, gpointer user_data)
 
 	if (!SOUP_MESSAGE_IS_STARTING (msg)) {
 		soup_message_queue_remove (priv->queue, item);
+
+		g_mutex_lock (priv->host_lock);
+		host = get_host_for_message (session, item->msg);
+		host->num_messages--;
+		g_mutex_unlock (priv->host_lock);
+
 		g_signal_handlers_disconnect_by_func (msg, message_finished, item);
 		/* g_signal_handlers_disconnect_by_func doesn't work if you
 		 * have a metamarshal, meaning it doesn't work with
@@ -1250,8 +1509,15 @@ queue_message (SoupSession *session, SoupMessage *msg,
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupMessageQueueItem *item;
+	SoupSessionHost *host;
 
 	item = soup_message_queue_append (priv->queue, msg, callback, user_data);
+
+	g_mutex_lock (priv->host_lock);
+	host = get_host_for_message (session, item->msg);
+	host->num_messages++;
+	g_mutex_unlock (priv->host_lock);
+
 	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
 
 	g_signal_connect_after (msg, "finished",
@@ -1496,6 +1762,38 @@ soup_session_abort (SoupSession *session)
 	g_slist_free (conns);
 }
 
+/**
+* soup_session_prepare_for_uri:
+* @session: a #SoupSession
+* @uri: a #SoupURI which may be required
+*
+* Tells @session that @uri may be requested shortly, and so the
+* session can try to prepare (resolving the domain name, obtaining
+* proxy address, etc.) in order to work more quickly once the URI is
+* actually requested.
+*
+* This method acts asynchronously, in @session's %async_context.
+* If you are using #SoupSessionSync and do not have a main loop running,
+* then you can't use this method.
+*
+* Since: 2.30
+**/
+void
+soup_session_prepare_for_uri (SoupSession *session, SoupURI *uri)
+{
+	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupSessionHost *host;
+	SoupAddress *addr;
+
+	g_mutex_lock (priv->host_lock);
+	host = get_host_for_uri (session, uri);
+	addr = g_object_ref (host->addr);
+	g_mutex_unlock (priv->host_lock);
+
+	soup_address_resolve_async (addr, priv->async_context,
+				    NULL, NULL, NULL);
+}
+
 /**
  * soup_session_add_feature:
  * @session: a #SoupSession
